@@ -156,9 +156,32 @@ end
 
 (** {2 General database operations} *)
 
-external db_open : string -> db = "caml_sqlite3_open"
-(** [db_open filename] opens the database file [filename], and returns
-    a database handle. *)
+val db_open :
+  ?mode : [ `READONLY | `NO_CREATE ] ->
+  ?mutex : [ `NO | `FULL ] ->
+  ?cache : [ `SHARED | `PRIVATE ] ->
+  ?vfs : string ->
+  string ->
+  db
+(** [db_open ?mode ?mutex ?cache ?vfs filename] opens the database file
+    [filename], and returns a database handle.
+
+    The optional arguments [mode] and [mutex] are only meaningful with SQlite
+    versions >= 3.5, [cache] only for versions >= 3.6.18.  For older versions an
+    exception will be raised if any of them is set to a non-default value.  The
+    database is opened read-only if [`READONLY] is passed as mode.  The database
+    file will not be created if it is missing and [`NO_CREATE] is set.  [mutex]
+    determines how the database is accessed.  The mutex parameters [`NO] and
+    [`FULL] correspond to [SQLITE_OPEN_NOMUTEX] and [SQLITE_OPEN_FULLMUTEX] in
+    the SQLite3 API respectively.  The cache parameters [`SHARED] and [`PRIVATE]
+    correspond to [SQLITE_OPEN_SHAREDCACHE] and [SQLITE_OPEN_PRIVATECACHE] in
+    the SQLite3 API respectively.
+
+    @param mode default = read-write, create
+    @param mutex default = nothing
+    @param cache default = nothing
+    @param vfs default = nothing
+*)
 
 external db_close : db -> bool = "caml_sqlite3_close"
 (** [db_close db] closes database [db] and invalidates the handle.
